#!/usr/bin/env perl
use strict;
use warnings;
use 5.10.0;

BEGIN {
	use Config;
	push @INC, '/opt/linode/longview';
	push @INC, '/opt/linode/longview/lib/perl5';
	push @INC, "/opt/linode/longview/lib/perl5/${Config{archname}}/";
	push @INC, '/opt/linode/longview/usr/include';
	$Net::HTTP::SOCKET_CLASS = 'IO::Socket::INET6';
	require Net::HTTP;
}

use JSON::PP;
use Try::Tiny;
use Sys::Hostname;
use LWP::UserAgent;
use Compress::Zlib;
use IO::Socket::INET6;
use Linode::Longview::DataGetter 'get';
use Linode::Longview::Util ':DRIVER';

$logger->info("Starting Longview Agent version $VERSION");

$logger->logdie('Longview must be run as root in order to collect data') unless ($< == 0);
my $pid = check_already_running();
$logger->logdie("The Longview agent is already running as PID: $pid") if $pid;

my $confdir    = '/etc/linode';
my $api_key_file = "$confdir/longview.key";

my $stats = {
	apikey  => scalar(slurp_file($api_key_file)),
	version => '1.0',
	payload => [],
};

unless ($stats->{apikey}){
	print "\nNo api key found. Please enter your API Key: ";
	chomp(my $key = <>);
	unless ($key =~ /^[0-9A-F]{8}-(?:[0-9A-F]{4}-){2}[0-9A-F]{16}\z$/){
		print "Invalid API Key\n";
		exit 1;
	}
	$stats->{apikey} = $key;
	umask 066;
	mkdir $confdir;
	open my $fh, '>', $api_key_file or $logger->logdie("Couldn't open $api_key_file for writing: $!");
	print $fh $key;
	close $fh or $logger->logdie("Couldn't close $api_key_file: $!");
}
$logger->logdie('Invalid API key') unless ($stats->{apikey} =~ /^[0-9A-F]{8}-(?:[0-9A-F]{4}-){2}[0-9A-F]{16}\z$/);
_prep_for_main();

my ($quit, $data) = (0, {});
while (!$quit) {
	my $sleep = $SLEEP_TIME;
	$data->{timestamp} = time;
	get($_,$data,) for (qw(Processes Ports Memory CPU Network Disk SysInfo Packages));

	constant_push($stats->{payload},$data);
	$data = {};

	$stats->{timestamp} = time;
	my $req = post($stats);

	if ($req->is_success){
		$logger->debug($req->status_line);
		$logger->debug($req->decoded_content);
		my $rep;
		try {
			$rep = decode_json($req->decoded_content);
		} catch {
			$logger->debug("Couldn't decode JSON response: $_");
		};
		$sleep = $rep->{sleep} if defined $rep->{sleep};
		if (defined($rep->{die}) && $rep->{die} eq 'please') {
			$logger->logdie('Server has requested this API Key stop sending data');
		}
		@{$stats->{payload}} = ();
	}
	else{
		$logger->info($req->status_line);
		$logger->info($req->decoded_content);
	}

	sleep $sleep;
}

sub _prep_for_main {
	chown 0, 0, $api_key_file;
	chmod 0600, $api_key_file;

	daemonize_self();
	enable_debug_logging() if(defined $ARGV[0] && $ARGV[0] eq 'Debug');

	$0 = 'linode-longview';
	$SIG{TERM} = $SIG{INT} = $SIG{HUP} = $SIG{QUIT} = sub { $quit = 1 };
	$logger->info('Start up complete');
}

