package Linode::Longview::DataGetter::Disk;

require 'syscall.ph';
use Linode::Longview::Util qw(:BASIC detect_system);
use POSIX 'uname';
use Cwd 'abs_path';
use Config;

sub get {
	my (undef, $dataref) = @_;
	$logger->trace('Collecting Disk info');

	my @swaps   = @{ _get_swap_info() };
	_get_mounted_info($dataref);

	# OpenVZ doesn't do /proc/diskstats, bail
	return $dataref if ( detect_system() eq 'openvz' );

	# get the information from /proc/diskstats
	@data = slurp_file( $PROCFS . 'diskstats' )or do {
		$logger->info("Couldn't read ${PROCFS}diststats: $!");
		return $dataref;
	};
	my %dev_mapper;
	for my $line (@data) {
		$line =~ s/^\s*//;

		#  202 0 xvda 3125353 13998 4980 2974 366 591 760 87320 15 366 9029
		my ( $major, $minor, $device, $reads, $writes ) = ( split( /\s+/, $line ) )[ 0..3, 7 ];
		$device = '/dev/' . $device;

		if (substr($device,0,8) eq '/dev/dm-') {
			# if the filesystem sees it under /dev
			if ( -b $device ) {
				unless (keys(%dev_mapper)) {
					%dev_mapper = map { substr(readlink($_),3) => substr($_,12); } (glob("/dev/mapper/*"));
				}
				if (exists($dev_mapper{substr($device,5)})) {
					$dataref->{INSTANT}->{"Disk.$device.label"} = $dev_mapper{substr($device,5)};
				}
			} else {
				unless (keys(%dev_mapper)) {
					%dev_mapper = map {
						my $rdev=(stat($_))[6];
						my $major_m = ($rdev & 03777400) >> 0000010;
						my $minor_m = ($rdev & 037774000377) >> 0000000;
						join('_', $major_m,$minor_m) => $_;
					} glob ("/dev/mapper/*");
				}
				if (exists($dev_mapper{"$major_$minor"})) {
					$dataref->{INSTANT}->{"Disk.$device.label"} = $dev_mapper{"$major_$minor"};
				}
			}
		} elsif ($device =~ m|(/dev/md\d+)(p\d+)?|) {
			if (defined($2)) {
				# it's a partition
				$dataref->{INSTANT}->{"Disk.$device.childof"} = $1;
				push @{$dataref->{INSTANT}->{"Disk.$1.children"}}, "$1$2";
			} else {
				next unless ($reads || $writes);
			}
		} else {
			# don't check out this guy if he has no writes and isn't someone we care about
			next unless ( (exists($dataref->{LONGTERM}->{"Disk.$device.fs.free"})) || ($device ~~ @swaps) );
		}
		$dataref->{LONGTERM}->{"Disk.$device.reads"}  = $reads + 0;
		$dataref->{LONGTERM}->{"Disk.$device.writes"} = $writes + 0;
		$dataref->{INSTANT}->{"Disk.$device.isswap"}  = $device ~~ @swaps ? 1 : 0;
	}

	return $dataref;
}

sub _get_swap_info {
	my @proc_swaps = slurp_file( $PROCFS . 'swaps' ) or do {
		$logger->info("Couldn't read ${PROCFS}swaps: $!");
		return [];
	};

	# first line of /proc/swaps is a header
	shift @proc_swaps;

	my @swaps;
	for my $line (@proc_swaps) {

		# /dev/xvdb partition 262140 122988 -1
		( $device, $type, $size, $used ) = ( split( /\s+/, $line ) )[ 0 .. 3 ];

		# not tracking file-backed swaps
		next if $type ne 'partition';
		push @swaps, $device;
	}
	return \@swaps;
}

sub _get_mounted_info {
	my $dataref = shift;
	my %devices;
	my @mtab = slurp_file('/etc/mtab') or do {
		$logger->info("Couldn't read /etc/mtab: $!");
		return $dataref;
	};
	for my $line ( grep {m|^/|} @mtab ) {
		my ( $device, $path ) = ( $line =~ m|^(\S+)\s+(\S+)| );
		next unless $device && $path;

		if ( $device =~ m|^/dev/mapper| ) {
			$device = abs_path( '/dev/mapper/' . readlink($device) );
		}

		if ($device =~ m|/dev/disk/by-uuid/[0-9a-f-]+|) {
			$device = abs_path( '/dev/' . readlink($device) );
		}

		if ($device eq '/dev/root') {
			my $root_dev = slurp_file($PROCFS . 'cmdline');
			($device) = ($root_dev =~ m|root=(\S+)|);
			if ($device =~ /UUID=([0-9a-f-]*)/) {
				$device = abs_path ('/dev/' . readlink("/dev/disk/by-uuid/$1"));
			}
		}

		my ( $type, $bsize, $blocks, $bfree, $bavail, $files, $ffree )
			= _statfs($path);

		$dataref->{LONGTERM}->{"Disk.$device.fs.free"}   = $bfree * $bsize;
		$dataref->{LONGTERM}->{"Disk.$device.fs.total"}  = $bsize * $blocks;
		$dataref->{LONGTERM}->{"Disk.$device.fs.ifree"}  = $ffree;
		$dataref->{LONGTERM}->{"Disk.$device.fs.itotal"} = $files;
		$dataref->{INSTANT}->{"Disk.$device.fs.path"}    = $path;
		$dataref->{INSTANT}->{"Disk.$device.mounted"}    = 1;
	}
	return $dataref;
}

sub _statfs {
	my $path = shift or do{
		$logger->info("Unable to stat $path: $!");
		return ();
	};
	my $fmt = '\0' x 512;
	syscall( &SYS_statfs, $path, $fmt ) == 0 or do {
		$logger->info("Couldn't call statfs syscall: $!");
		return ();
	};

	# unsigned long vs unsigned quad on 32 vs 64 bit boxes
	if ( $Config{alignbytes} eq '4' ) {
		return unpack 'L7', $fmt;
	}
	elsif ( $Config{alignbytes} eq '8' ) {
		return unpack 'Q7', $fmt;
	}
	else {
		$logger->info('Unable to determine architecture');
		return ();
	}
}

1;
