#!/usr/bin/perl

package Linode::Longview::DataGetter::Processes;

use POSIX;
use Linode::Longview::Util qw(:BASIC $TICKS);

use Exporter 'import';
our @EXPORT_OK = (qw(process_info process_list));

sub process_list {
	my @procs = map { s|$PROCFS||; $_ } glob "${PROCFS}[0-9]*";
	return @procs;
}

sub process_info {
	my $pid = shift;
	my %proc;

	# all or nothing, grab what we need asap
	my @proc_file  = slurp_file( $PROCFS . "$pid/status" ) or return;
	my $stat_line = slurp_file( $PROCFS . "$pid/stat" )   or return;
	# some kernels don't have /proc/pid/io, let the for loop fall apart later
	my @iosample  = slurp_file( $PROCFS . "$pid/io" ) ;

	for my $line (@proc_file) {
		if ( $line =~ m/^Name:\s+(.*)/ )         { $proc{name} = $1; }
		if ( $line =~ m/^Pid:\s+(.*)/ )          { $proc{pid}  = $1; }
		if ( $line =~ m/^PPid:\s+(.*)/ )         { $proc{ppid} = $1; }
		if ( $line =~ m/^Uid:\s+.*?\s+(.*?)\s/ ) { $proc{uid}  = $1; }
		if ( $line =~ m/^VmRSS:\s+(.*)\s+kB/ )   { $proc{mem}  = $1; }
		last if ( $line =~ m/^Threads/ );
	}

	$proc{user}  = ( getpwuid( $proc{uid} ) )[0];
	$proc{group} = ( getgrgid( $proc{gid} ) )[0];

	# cpu
	my ( $user_time, $system_time, $start_jiffies )
		= ( split /\s+/, $stat_line )[ 13, 14, 21 ];
	$proc{cpu} = $user_time + $system_time;
	$proc{age} = proc_age($start_jiffies);

	# io
	for $line (@iosample) {
		$proc{ioreadkbytes}  = $1 if ( $line =~ m/^read_bytes:\s+(.*)/ );
		$proc{iowritekbytes} = $1 if ( $line =~ m/^write_bytes:\s+(.*)/ );
	}

	return %proc;
}

sub get {
	my (undef, $dataref) = @_;

	$logger->trace('Collecting Process information');

	my @procs = process_list();
	for my $proc (@procs) {
		my %info = process_info($proc);
		next if ( !%info );    #No such proc, it probably died while we were gathering info
		next if ( $info{ppid} == 2 );
		next if ( $info{pid} == 2 );
		$info{name} =~ s/\./\\\./g;

		$dataref->{LONGTERM}->{"Processes.${info{name}}.${info{user}}.count"}++;
		$dataref->{LONGTERM}->{"Processes.${info{name}}.${info{user}}.mem"} 			+= $info{mem} + 0;
		$dataref->{LONGTERM}->{"Processes.${info{name}}.${info{user}}.cpu"} 			+= $info{cpu}			if (defined $info{cpu});
		$dataref->{LONGTERM}->{"Processes.${info{name}}.${info{user}}.ioreadkbytes"}	+= $info{ioreadkbytes}	if (defined $info{ioreadkbytes});
		$dataref->{LONGTERM}->{"Processes.${info{name}}.${info{user}}.iowritekbytes"}	+= $info{iowritekbytes}	if (defined $info{iowritekbytes});
		$dataref->{LONGTERM}->{"Processes.${info{name}}.${info{user}}.zz_age"} 			 = $info{age}
			if ((!defined($dataref->{LONGTERM}->{"Processes.${info{name}}.${info{user}}.zz_age"} )) ||
				$info{age} > $dataref->{LONGTERM}->{"Processes.${info{name}}.${info{user}}.zz_age"} );
	}

	for my $name (sort keys %{$dataref->{LONGTERM}}) {
		next unless $name =~ /^Processes/;
		if($name =~ /zz_age$/ ){
			delete $dataref->{LONGTERM}->{$name};
			next;
		}
		my $age_key = $name;
		$age_key =~ s/[^\.]*$/zz_age/;
		delete $dataref->{LONGTERM}->{$name} if ($dataref->{LONGTERM}->{$age_key} < 60);
	}

	return $dataref;
}

sub proc_age {
	my $start_jiffies   = shift;
	my $uptime = ( split( /\s+/, slurp_file( $PROCFS . 'uptime' ) ) )[0];
	my $current_jiffies = $uptime * $TICKS;
	return ($current_jiffies - $start_jiffies) / $TICKS;
}

1;
